// GLMeshbuilderx
{: Creation of true meshes for standard GLScene Objects.
   Created by : Graham Kennedy
   Based on original code by  Joen Joensen.

   Contributed to the GLScene community.

   At the moment I've created a rather clunky 'BuildFreeFormFrom'
   which will create a freeform from the GLScene Object passed to it (for those objects
   listed as supported) as well as copying the relevant attributes required to clone it.

   It would be really great to see these as actual procedures off each of the standard objects
   so that the relevant meshobject can be kept in step with any changes made in future.
<p>

	<b>History : </b><font size=-1><ul>
      <li>05/04/04 - GAK - Created, based on GLMeshbuilder, kept separate in case of problems
          Version is currently very rough, distinct problems around the normals (as you can see
          I've pretty much just lifted the buildlist of the original object in many cases).

          Currently covers basic geometry :
            GLCube
            GLSphere
            GLPlane
            GLFrustrum
            GLCylinder
            GLCone
            GLDisk
            GLDodecahedron

   </ul></font>
}

unit GLMeshBuilderx;

interface
uses
  GLScene, GLObjects,GLVectorFileObjects,VectorGeometry, GLMisc,OpenGL1x,GLGeomObjects,
  GLPolyhedron,
  dialogs, sysutils,

  VectorLists,
  MeshUtils,
  PersistentClasses;

  // base 'builder' routines, will generate the given object within the meshobject passed to it
  procedure BuildMeshCube(mo : tmeshobject;width,height,depth:single;parts:TCubeParts);
  procedure BuildMeshSphere(mo : tmeshobject;radius:single;stacks,slices,start,stop,top,bottom:integer;topcap,bottomcap:TCapType);
  procedure BuildMeshPlane(mo : tmeshobject;width,height:single;xtiles,ytiles:integer;xoffset,xscope,yoffset,yscope:single;style:TPlaneStyles);
  procedure BuildMeshFrustrum(mo : tmeshobject;height,apexheight,basedepth,basewidth:single;parts:TFrustrumParts);
  procedure BuildMeshCylinder(mo : tmeshobject;topradius,bottomradius,height:single;stacks,slices:integer;parts:TCylinderParts);
  procedure BuildMeshDodecahedron(mo:tmeshobject);
  procedure BuildMeshDisk(mo:tmeshobject;innerradius,outerradius,startangle,sweepangle:single;loops,slices:integer);

  procedure BuildMeshTorus(mo : tmeshobject;MinorRadius,MajorRadius:single);

  // generic routine to create a mesh based copy of a GLScene object
  function BuildFreeFormFrom(source:TGLSceneObject):TGLFreeForm;

  procedure AddToFreeForm(source:TGLSceneObject; var GLFreeForm:TGLFreeForm);
  procedure StripifyMeshObject(source:TGLSceneObject; MeshObject: TMeshObject; GLFreeForm: TGLFreeForm);

implementation

procedure AddToFreeForm(source: TGLSceneObject; var GLFreeForm:TGLFreeForm);
var
  MeshObject : TMeshObject;
  CylinderParts: TCylinderParts;
  i, j: integer;
  FaceGroup: TFGVertexIndexList;
Begin
  MeshObject:= TMeshObject.CreateOwned(GLFreeForm.MeshObjects);
  MeshObject.Mode:= momFaceGroups;

  if source is TGLCube then
  begin
      with source as tglcube do
      begin
        BuildMeshCube(MeshObject, cubewidth,cubeheight, CubeDepth,parts);
      end;
  end

  else if source is TGLsphere then
  begin
      with source as tglsphere do
      begin
        BuildMeshsphere(MeshObject, Radius,Stacks,Slices,Start,Stop,top,bottom,topcap,bottomcap);
        // temp fix, until I work out why the radius is not working in the creation routine
        MeshObject.Vertices.Scale(0.5);
        //ff.scale.X := ff.scale.x*radius;
        //ff.scale.y := ff.scale.y*radius;
        //ff.scale.z := ff.scale.z*radius;
      end;
  end

  else if source is tglplane then
  begin
    with source as tglplane do
    begin
      buildmeshplane(MeshObject, width,height,xtiles,ytiles,xoffset,xscope,yoffset,yscope,style);
    end;
  end

  else if source is tglfrustrum then
  begin
    with source as tglfrustrum do
    begin
      buildmeshfrustrum(MeshObject, height,ApexHeight,BaseDepth,BaseWidth,parts);
    end;
  end

  else if source is tglcone then
  begin
    with source as tglcone do
    begin
      CylinderParts:= [];
      if cosides in parts then CylinderParts:= CylinderParts + [cySides];
      if cobottom in parts then CylinderParts:= CylinderParts + [cyBottom];
//      buildmeshcone( mo,BottomRadius,Height,Stacks,Slices,parts);
      buildmeshcylinder(MeshObject, 0,BottomRadius,Height,Stacks,Slices, CylinderParts);
    end;
  end

  else if source is tglcylinder then
  begin
    with source as tglcylinder do
    begin
      buildmeshcylinder(MeshObject, topradius,BottomRadius,Height,Stacks,Slices,parts);
    end;
  end

  else if source is TGLDodecahedron then
  begin
    with source as TGLDodecahedron do
    begin
      buildmeshdodecahedron(MeshObject);
    end;
  end

  else if source is TGLTorus then
  begin
    with source as TGLTorus do
    begin
      BuildMeshTorus(MeshObject, MinorRadius, MajorRadius);
    end;
  end

  else if source is TGLdisk then
  begin
    with source as TGLdisk do
    begin
      buildmeshdisk(MeshObject, InnerRadius,OuterRadius,StartAngle,SweepAngle,Loops,Slices);
    end;
  end;

  MeshObject.Vertices.TransformAsPoints(source.AbsoluteMatrix);
  StripifyMeshObject(source, MeshObject, GLFreeForm);
  GLFreeForm.StructureChanged;
end;

procedure StripifyMeshObject(source:TGLSceneObject; MeshObject: TMeshObject; GLFreeForm: TGLFreeForm);
var
  i, j: Integer;
  list, normals: TAffineVectorList;
  indices: TIntegerList;
  strips: TPersistentObjectList;
  NewMeshObject: TMeshObject;
  FaceGroup: TFGVertexIndexList;
begin
  list:= MeshObject.ExtractTriangles;
  indices:= BuildVectorCountOptimizedIndices(list);
  RemapAndCleanupReferences(list, indices);
  normals:= BuildNormals(list, indices);
  strips:= StripifyMesh(indices, list.Count, True);
  MeshObject.Clear;

  j:= GLFreeForm.MaterialLibrary.Materials.Count;
  with GLFreeForm.MaterialLibrary.Materials.Add do begin
    Name:=IntToStr(j);
    //Material.FrontProperties.Ambient.Color:= VectorMake(0.2, 0.2, 0.2);//clrGray20;
    Material.FrontProperties.Diffuse.Color:= source.Material.FrontProperties.Diffuse.Color;//moColor;//clrGray80
    Material.FrontProperties.Emission.Color:= VectorMake(0.2, 0.2, 0.2);//clrBlack;

    Material.BlendingMode:= source.Material.BlendingMode;
    Material.FrontProperties.Diffuse.Alpha:= source.Material.FrontProperties.Diffuse.Alpha;
    //Material:= source.Material;
    end;

  NewMeshObject:= TMeshObject.CreateOwned(GLFreeForm.MeshObjects);
  NewMeshObject.Vertices:= list;
  NewMeshObject.Mode:= momFaceGroups;
  NewMeshObject.Normals:= normals;
  for i:= 0 to strips.Count - 1 do begin
    FaceGroup:= TFGVertexIndexList.CreateOwned(NewMeshObject.FaceGroups);
    FaceGroup.VertexIndices:= (strips[i] as TIntegerList);
    if i > 0
      then FaceGroup.Mode:= fgmmTriangleStrip
      else FaceGroup.Mode:= fgmmTriangles;
    FaceGroup.MaterialName:= inttostr(j);
  end;
  GLFreeForm.StructureChanged;
  strips.CleanFree;
  indices.Free;
  normals.Free;
  list.Free;
end;











function BuildFreeFormFrom(source:TGLSceneObject):TGLFreeForm;
var
  ff : tglfreeform;
  mo : tmeshobject;
  cp : TCylinderParts;
Begin
  ff := TGLFreeForm.Create(source.Owner);
  ff.Parent := source.Parent;
  ff.Position.Assign(source.Position);
  ff.Scale.Assign(source.Scale);
  ff.Up.assign(source.Up);
  ff.Material.Assign(source.material);
  ff.Direction.Assign(source.direction);

  mo :=TMeshObject.CreateOwned(ff.MeshObjects);
  mo.Mode := momFaceGroups;

  if source is TGLCube then
  begin
      with source as tglcube do
      begin
        BuildMeshCube(mo,cubewidth,cubeheight,CubeDepth,parts);
      end;
  end;

  if source is TGLsphere then
  begin
      with source as tglsphere do
      begin
        BuildMeshsphere(mo,Radius,Stacks,Slices,Start,Stop,top,bottom,topcap,bottomcap);
        // temp fix, until I work out why the radius is not working in the creation routine
        ff.scale.X := ff.scale.x*radius;
        ff.scale.y := ff.scale.y*radius;
        ff.scale.z := ff.scale.z*radius;
      end;
  end;

  if source is tglplane then
  begin
    with source as tglplane do
    begin
      buildmeshplane(mo,width,height,xtiles,ytiles,xoffset,xscope,yoffset,yscope,style);
    end;
  end;

  if source is tglfrustrum then
  begin
    with source as tglfrustrum do
    begin
      buildmeshfrustrum( mo,height,ApexHeight,BaseDepth,BaseWidth,parts);
    end;
  end;

  if source is tglcone then
  begin
    with source as tglcone do
    begin
      cp := [];
      if cosides in parts then cp := cp + [cySides];
      if cobottom in parts then cp := cp + [cyBottom];
//      buildmeshcone( mo,BottomRadius,Height,Stacks,Slices,parts);
      buildmeshcylinder( mo,0,BottomRadius,Height,Stacks,Slices,cp);
    end;
  end;

  if source is tglcylinder then
  begin
    with source as tglcylinder do
    begin
      buildmeshcylinder( mo,topradius,BottomRadius,Height,Stacks,Slices,parts);
    end;
  end;

  if source is TGLDodecahedron then
  begin
    with source as TGLDodecahedron do
    begin
      buildmeshdodecahedron(mo);
    end;
  end;

  if source is TGLdisk then
  begin
    with source as TGLdisk do
    begin
      buildmeshdisk(mo,InnerRadius,OuterRadius,StartAngle,SweepAngle,Loops,Slices);
    end;
  end;

  ff.StructureChanged;

  result :=ff;
end;

procedure BuildMeshCube(mo:tmeshobject;width,height,depth:single;parts:TCubeParts);
var
  fg : TFGVertexIndexList;
  w,h,d:single;
  fv : integer;


  procedure AddFaceGroup(st:integer);
  begin
     fg.add(st);
     fg.add(st+1);
     fg.add(st+2);
     fg.add(st+2);
     fg.add(st+3);
     fg.add(st);
  end;

begin
  w := width/2.0;
  h := height/2.0;
  d := depth/2.0;
  FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);


  // front
  if cpFront in Parts then
  begin
    mo.Normals.add(0,0,d);
    fv := mo.Vertices.add(-w,-h, d);
        mo.Vertices.add( w,-h, d);
        mo.Vertices.add( w, h, d);
        mo.Vertices.add(-w, h, d);
        mo.TexCoords.Add(0, 0);
        mo.TexCoords.Add(1, 0);
      mo.TexCoords.Add(1, 1);
      mo.TexCoords.Add(0, 1);
    AddFaceGroup(fv);
  end;

  // back
  if cpback in Parts then
  begin
    mo.Normals.add(0,0,-d);
    fv := mo.Vertices.add(-w,-h,-d);
        mo.Vertices.add(-w, h,-d);
        mo.Vertices.add( w, h,-d);
        mo.Vertices.add( w,-h,-d);
        mo.TexCoords.Add(1, 0);
        mo.TexCoords.Add(1, 1);
        mo.TexCoords.Add(0, 1);
        mo.TexCoords.Add(0, 0);
    AddFaceGroup(fv);
  end;

  // top
  if cptop in Parts then
  begin
    mo.Normals.add(0,d,0);
    mo.Normals.add(0,d,0);
    fv := mo.Vertices.add(-w, h,-d);
        mo.Vertices.add(-w, h, d);
        mo.Vertices.add( w, h, d);
        mo.Vertices.add( w, h,-d);
        mo.TexCoords.Add(0, 1);
        mo.TexCoords.Add(0, 0);
        mo.TexCoords.Add(1, 0);
        mo.TexCoords.Add(1, 1);
    AddFaceGroup(fv);
  end;

  // bottom
  if cpbottom in Parts then
  begin
    mo.Normals.add(0,-d,0);
    fv := mo.Vertices.add(-w,-h,-d);
        mo.Vertices.add( w,-h,-d);
        mo.Vertices.add( w,-h, d);
        mo.Vertices.add(-w,-h, d);
        mo.TexCoords.Add(1, 1);
        mo.TexCoords.Add(0, 1);
        mo.TexCoords.Add(0, 0);
        mo.TexCoords.Add(1, 0);
    AddFaceGroup(fv);
  end;

  // right
  if cpright in Parts then
  begin
    mo.Normals.add(d,0,d);
    fv := mo.Vertices.add( w,-h,-d);
        mo.Vertices.add( w, h,-d);
        mo.Vertices.add( w, h, d);
        mo.Vertices.add( w,-h, d);
        mo.TexCoords.Add(1, 0);
        mo.TexCoords.Add(1, 1);
        mo.TexCoords.Add(0, 1);
        mo.TexCoords.Add(0, 0);
    AddFaceGroup(fv);
  end;

  // left
  if cpleft in Parts then
  begin
    mo.Normals.add(-d,0,0);
    fv := mo.Vertices.add(-w,-h,-d);
        mo.Vertices.add(-w,-h, d);
        mo.Vertices.add(-w, h, d);
        mo.Vertices.add(-w, h,-d);
        mo.TexCoords.Add(0, 0);
        mo.TexCoords.Add(1, 0);
        mo.TexCoords.Add(1, 1);
        mo.TexCoords.Add(0, 1);
    AddFaceGroup(fv);
  end;
end;


procedure BuildMeshSphere(mo:tmeshobject;radius:single;stacks,slices,start,stop,top,bottom:integer;topcap,bottomcap:TCapType);
var
  fg : TFGVertexIndexList;
  fv1,cv1,pv1 : integer;
  fv2,cv2,pv2 : integer;

   V1, V2, N1 : TAffineVector;
   AngTop, AngBottom, AngStart, AngStop, StepV, StepH : Extended;
   SinP, CosP, SinP2, CosP2, SinT, CosT, Phi, Phi2, Theta: Extended;
   uTexCoord, uTexFactor, vTexFactor, vTexCoord0, vTexCoord1 : Single;
   I, J: Integer;
   DoReverse: Boolean;
begin
  FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);

   // common settings
   AngTop:=DegToRad(top);
   AngBottom:=DegToRad(Bottom);
   AngStart:=DegToRad(Start);
   AngStop:=DegToRad(Stop);
   StepH:=(AngStop - AngStart) / Slices;
   StepV:=(AngTop - AngBottom) / Stacks;

   if (Top < 90) and (TopCap in [ctCenter, ctFlat]) then
   begin
      SinCos(AngTop, SinP, CosP);
      sinp := sinp;// * radius;
      cosp := cosp;// * radius;
      mo.TexCoords.Add(0.5,0.5);

//      xglTexCoord2f(0.5, 0.5);

      if DoReverse then
         mo.Normals.Add(0,-1,0)//  glNormal3f(0, -1, 0)
      else mo.Normals.add(0,1,0);  //glNormal3f(0, 1, 0);

      mo.Normals.Add(0,1,0); //??
      if TopCap = ctCenter then
        fv1 := mo.Vertices.Add(0,0,0)
      else
      begin
         fv1 := mo.vertices.add(0,sinp,0);
         N1:=YVector;

         if DoReverse then
            N1[1]:=-N1[1];
      end;

      V1[1]:=SinP;//*radius;
      Theta:=AngStart;
      pv1 := -1;
      cv1 := -1;
      for I:=0 to Slices do
      begin
         SinCos(Theta, SinT, CosT);
         V1[0]:=CosP*SinT;//*radius;
         V1[2]:=CosP*CosT;//*radius;
         if TopCap=ctCenter then
         begin
            N1:=VectorPerpendicular(YVector, V1);
            //if DoReverse then NegateVector(N1);
         end;
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);
         if pv1 <> -1 then
         begin
            fg.Add(fv1);
            fg.Add(pv1);
            fg.add(cv1);
         end;
         mo.Normals.Add(n1);
         mo.TexCoords.Add(SinT*0.5+0.5, CosT*0.5+0.5);

//         xglTexCoord2f(SinT*0.5+0.5, CosT*0.5+0.5);
//         glNormal3fv(@N1);
//         glVertex3fv(@V1);
         Theta:=Theta + StepH;
      end;
//      glEnd;

   end;


   // main body
   Phi:=AngTop;
   Phi2:=Phi-StepV;
   uTexFactor:=1/Slices;
   vTexFactor:=1/Stacks;

   for J:=0 to Stacks-1 do begin
      Theta:=AngStart;
      SinCos(Phi, SinP, CosP);
      SinCos(Phi2, SinP2, CosP2);
      V1[1]:=SinP;
      V2[1]:=SinP2;
      vTexCoord0:=1-j*vTexFactor;
      vTexCoord1:=1-(j+1)*vTexFactor;

      pv1 := -1;
      cv1 := -1;
      pv2 := -1;
      cv2 := -1;
      for i:=0 to Slices do begin

         SinCos(Theta, SinT, CosT);
         V1[0]:=CosP * SinT;
         V2[0]:=CosP2 * SinT;
         V1[2]:=CosP * CosT;
         V2[2]:=CosP2 * CosT;

         uTexCoord:=i*uTexFactor;
//         xglTexCoord2f(uTexCoord, vTexCoord0);
         mo.TexCoords.Add(uTexCoord, vTexCoord0);
         if DoReverse then begin
            N1:=VectorNegate(V1);
//            glNormal3fv(@N1);
            mo.Normals.Add(n1);
         end
         else
         begin
//          glNormal3fv(@V1);
            mo.Normals.Add(v1);
         end;
//         glVertex3fv(@V1);
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);

//         xglTexCoord2f(uTexCoord, vTexCoord1);
         mo.TexCoords.Add(uTexCoord, vTexCoord1);
         if DoReverse then
         begin
            N1:=VectorNegate(V2);
//            glNormal3fv(@N1);
            mo.Normals.Add(n1);
         end
         else
         begin
//          glNormal3fv(@V2);
            mo.Normals.Add(v2);
         end;
         pv2 := cv2;
         cv2 := mo.Vertices.Add(v2);
//         mo.Normals.Add(v2);
//         glVertex3fv(@V2);

         if pv1 <> -1 then
         begin
            fg.Add(pv1);
            fg.Add(cv2);
            fg.add(cv1);
            fg.Add(cv2);
            fg.Add(pv1);
            fg.add(pv2);
         end;

         Theta:=Theta+StepH;
      end;
      Phi:=Phi2;
      Phi2:=Phi2 - StepV;
   end;

   // bottom cap
   if (Bottom > -90) and (BottomCap in [ctCenter, ctFlat]) then
   begin
      SinCos(AngBottom, SinP, CosP);
//      xglTexCoord2f(0.5, 0.5);
      mo.TexCoords.add(0.5,0.5);
      if DoReverse then
         mo.Normals.add(0,1,0)//   glNormal3f(0, 1, 0)
      else mo.Normals.add(0,-1,0);//   glNormal3f(0, -1, 0);
      if BottomCap = ctCenter then
         fv1 := mo.Vertices.add(0,0,0) //glVertex3f(0, 0, 0)
      else
      begin
         fv1 := mo.Vertices.add(0,sinp,0);//glVertex3f(0, SinP, 0);
         if DoReverse then
            MakeVector(N1, 0, -1, 0)
         else N1:=YVector;
      end;
      V1[1]:=SinP;
      Theta:=AngStop;
      cv1 := -1;
      pv1 := -1;
      for I:=0 to Slices do begin
         SinCos(Theta, SinT, CosT);
         V1[0]:=CosP * SinT;
         V1[2]:=CosP * CosT;
         if TopCap = ctCenter then
         begin
            N1:=VectorPerpendicular(AffineVectorMake(0, -1, 0), V1);
            if DoReverse then NegateVector(N1);
         end;
         //xglTexCoord2f(SinT*0.5+0.5, CosT*0.5+0.5);
         mo.TexCoords.Add(SinT*0.5+0.5, CosT*0.5+0.5);
         mo.Normals.add(n1);
         //glNormal3fv(@N1);
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);
         //glVertex3fv(@V1);
         if pv1 <> -1 then
         begin
            fg.Add(fv1);
            fg.Add(pv1);
            fg.add(cv1);
         end;

         Theta:=Theta - StepH;
      end;
   end;

end;

procedure BuildMeshPlane(mo : tmeshobject;width,height:single;xtiles,ytiles:integer;xoffset,xscope,yoffset,yscope:single;style:TPlaneStyles);
var
  fg : TFGVertexIndexList;
  fv1,cv1,pv1 : integer;
  fv2,cv2,pv2 : integer;

   hw, hh, posXFact, posYFact, pX, pY0, pY1 : TGLFloat;
   tx0, tx1, ty0, ty1, texSFact, texTFact : TGLFloat;
   texS, texT0, texT1 : TGLFloat;
   x, y : Integer;
begin
   FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);

   hw:=Width*0.5;
   hh:=Height*0.5;

   mo.Normals.Add(zvector);
//   glNormal3fv(@ZVector);
   // determine tex coords extents
   if psTileTexture in Style then
   begin
      tx0:=XOffset;
      tx1:=XTiles*XScope+XOffset;
      ty0:=YOffset;
      ty1:=YTiles*YScope+YOffset;
   end
   else
   begin
      tx0:=0;
      ty0:=tx0;
      tx1:=XScope;
      ty1:=YScope;
   end;

//   if NoZWrite then
//      glDepthMask(False);

   if psSingleQuad in Style then
   begin
      // single quad plane
//      glBegin(GL_QUADS);
      mo.TexCoords.Add(tx1,ty1);
//         xglTexCoord2f(tx1, ty1);
      mo.Vertices.add(hw,hh);
//         glVertex2f( hw, hh);
      mo.TexCoords.Add(tx0,ty1);
//         xglTexCoord2f(tx0, ty1);
      mo.Vertices.add(-hw,hh);
//         glVertex2f(-hw, hh);
      mo.TexCoords.Add(tx0,ty0);
//         xglTexCoord2f(tx0, ty0);
      mo.Vertices.add(-hw,-hh);
//         glVertex2f(-hw, -hh);
      mo.TexCoords.Add(tx1,ty0);
//         xglTexCoord2f(tx1, ty0);
      mo.Vertices.add(hw,-hh);
//         glVertex2f( hw, -hh);
//      glEnd;
      fg.Add(1);
      fg.Add(2);
      fg.Add(0);
      fg.Add(0);
      fg.Add(2);
      fg.Add(3);
   end
   else
   begin
      // multi-quad plane (actually built from tri-strips)
      texSFact:=(tx1-tx0)/XTiles;
      texTFact:=(ty1-ty0)/YTiles;
      posXFact:=Width/XTiles;
      posYFact:=Height/YTiles;
      texT0:=0;
      pY0:=-hh;
      for y:=0 to YTiles-1 do
      begin
         texT1:=(y+1)*texTFact;
         pY1:=(y+1)*posYFact-hh;
//         glBegin(GL_TRIANGLE_STRIP);
         cv1 := -1;
         cv2 := -1;
         for x:=0 to XTiles do
         begin
            texS:=tx0+x*texSFact;
            pX:=x*posXFact-hw;
            mo.TexCoords.add(texs,text1);
//            xglTexCoord2f(texS, texT1);
            pv1 := cv1;
            cv1 := mo.Vertices.Add(px,py1);
//            glVertex2f(pX, pY1);
            mo.TexCoords.add(texs,text0);
//            xglTexCoord2f(texS, texT0);
            pv2 := cv2;
            cv2 := mo.Vertices.add(px,py0);
//            glVertex2f(pX, pY0);

            if pv1 <> -1 then
            begin
              fg.Add(pv1);
              fg.Add(cv2);
              fg.add(cv1);
              fg.Add(cv2);
              fg.Add(pv1);
              fg.add(pv2);
            end;

         end;
//         glEnd;
         texT0:=texT1;
         pY0:=pY1;
      end;
   end;
end;

procedure BuildMeshFrustrum(mo : tmeshobject;height,apexheight,basedepth,basewidth:single;parts:TFrustrumParts);
var
  fg : TFGVertexIndexList;
  fv1,cv1,pv1 : integer;

  HBW, HBD: TGLFloat; // half of width, half of depth at base
  HTW, HTD: TGLFloat; // half of width, half of depth at top of frustrum
  Sign: TGLFloat;     // +1 or -1
  Angle: TGLFloat;    // in radians
  ASin, ACos: TGLFloat;
begin
   FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);


//  if FNormalDirection = ndInside then
//    Sign := -1
//  else
    Sign := 1;
  HBW := BaseWidth * 0.5;
  HBD := BaseDepth * 0.5;
  HTW := HBW * (ApexHeight - Height) / ApexHeight;
  HTD := HBD * (ApexHeight - Height) / ApexHeight;

//  glBegin(GL_QUADS);

  if [fpFront, fpBack] * Parts <> [] then
  begin
    Angle := Arctan(ApexHeight/ HBD); // angle of front plane with bottom plane
    SinCos(Angle, ASin, ACos);
    if fpFront in Parts then
    begin
      mo.Normals.Add(0,Sign * ACos, Sign * ASin);
      mo.TexCoords.Add(xytexpoint); fv1 := mo.Vertices.add( HTW, Height, HTD);
      mo.TexCoords.Add(ytexpoint);         mo.Vertices.add(-HTW, Height, HTD);
      mo.TexCoords.Add(nulltexpoint);      mo.Vertices.add(-HBW, 0, HBD);
      mo.TexCoords.Add(xtexpoint);         mo.Vertices.add( HBW, 0, HBD);
      fg.Add(fv1);
      fg.Add(fv1+1);
      fg.Add(fv1+2);
      fg.Add(fv1+2);
      fg.Add(fv1+3);
      fg.Add(fv1+0);
    end;

    if fpBack in Parts then
    begin
      mo.Normals.Add(0, Sign * ACos, -Sign * ASin);
      mo.TexCoords.Add(ytexpoint); fv1 := mo.Vertices.add( HTW, Height,-HTD);
      mo.TexCoords.Add(nulltexpoint);     mo.Vertices.add( HBW, 0,-HBD);
      mo.TexCoords.Add(xtexpoint);        mo.Vertices.add(-HBW, 0,-HBD);
      mo.TexCoords.Add(xytexpoint);       mo.Vertices.add(-HTW, Height,-HTD);
      fg.Add(fv1);
      fg.Add(fv1+1);
      fg.Add(fv1+2);
      fg.Add(fv1+2);
      fg.Add(fv1+3);
      fg.Add(fv1+0);
    end;
  end;

  if [fpLeft, fpRight] * Parts <> [] then
  begin
    Angle := Arctan(ApexHeight/ HBW); // angle of side plane with bottom plane
    SinCos(Angle, ASin, ACos);
    if fpLeft in Parts then
    begin
      mo.Normals.Add(-Sign * ASin, Sign * ACos, 0);
      mo.TexCoords.Add(xytexpoint); fv1 := mo.Vertices.add(-HTW, Height, HTD);
      mo.TexCoords.Add(ytexpoint);         mo.Vertices.add(-HTW, Height,-HTD);
      mo.TexCoords.Add(nulltexpoint);      mo.Vertices.add(-HBW, 0,-HBD);
      mo.TexCoords.Add(xtexpoint);         mo.Vertices.add(-HBW, 0, HBD);
      fg.Add(fv1);
      fg.Add(fv1+1);
      fg.Add(fv1+2);
      fg.Add(fv1+2);
      fg.Add(fv1+3);
      fg.Add(fv1+0);
    end;

    if fpRight in Parts then
    begin
      mo.Normals.Add(Sign * ASin, Sign * ACos, 0);
      mo.TexCoords.Add(ytexpoint); fv1 := mo.Vertices.add( HTW, Height,HTD);
      mo.TexCoords.Add(nulltexpoint);     mo.Vertices.add( HBW, 0,HBD);
      mo.TexCoords.Add(xtexpoint);        mo.Vertices.add( HBW, 0,-HBD);
      mo.TexCoords.Add(xytexpoint);       mo.Vertices.add( HTW, Height,-HTD);
      fg.Add(fv1);
      fg.Add(fv1+1);
      fg.Add(fv1+2);
      fg.Add(fv1+2);
      fg.Add(fv1+3);
      fg.Add(fv1+0);

    end;
  end;

  if (fpTop in Parts) and (Height < ApexHeight) then
  begin
      mo.Normals.Add(0, Sign, 0);
      mo.TexCoords.Add(ytexpoint);    fv1 := mo.Vertices.add(-HTW, Height,-HTD);
      mo.TexCoords.Add(nulltexpoint);        mo.Vertices.add(-HTW, Height, HTD);
      mo.TexCoords.Add(xtexpoint);           mo.Vertices.add( HTW, Height, HTD);
      mo.TexCoords.Add(xytexpoint);          mo.Vertices.add( HTW, Height,-HTD);
      fg.Add(fv1);
      fg.Add(fv1+1);
      fg.Add(fv1+2);
      fg.Add(fv1+2);
      fg.Add(fv1+3);
      fg.Add(fv1+0);
  end;

  if fpBottom in Parts then
  begin
      mo.Normals.Add(0, -Sign, 0);
      mo.TexCoords.Add(nulltexpoint);  fv1 := mo.Vertices.add(-HBW, 0,-HBD);
      mo.TexCoords.Add(xtexpoint);            mo.Vertices.add( HBW, 0,-HBD);
      mo.TexCoords.Add(xytexpoint);           mo.Vertices.add( HBW, 0, HBD);
      mo.TexCoords.Add(ytexpoint);            mo.Vertices.add(-HBW, 0, HBD);
      fg.Add(fv1);
      fg.Add(fv1+1);
      fg.Add(fv1+2);
      fg.Add(fv1+2);
      fg.Add(fv1+3);
      fg.Add(fv1+0);

  end;


end;


procedure BuildMeshCylinder(mo : tmeshobject;topradius,bottomradius,height:single;stacks,slices:integer;parts:TCylinderParts);
var
  fg : TFGVertexIndexList;
  fv1,cv1,pv1 : integer;
  fv2,cv2,pv2 : integer;
  angstart,angstop,steph,stepv : extended;
  h : single;
  V1, V2, N1 : TAffineVector;
  i,j:integer;
  SinP, CosP, SinP2, CosP2, SinT, CosT, Phi, Phi2, Theta: Extended;
  uTexCoord, uTexFactor, vTexFactor, vTexCoord0, vTexCoord1 : Single;
  y,rad1,rad2:extended;
begin
   FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);
   h := height /2.0;

   AngStart:=DegToRad(0);
   AngStop:=DegToRad(360);
   StepH:=(AngStop - AngStart) / Slices;
   StepV:=height / Stacks;

  if cysides in parts then
  begin
   Phi:=h;
   Phi2:=Phi-StepV;
   uTexFactor:=1/Slices;
   vTexFactor:=1/Stacks;

   for J:=0 to Stacks-1 do begin
      Theta:=AngStart;
      rad1 := topradius+((bottomradius-topradius)/stacks)*j;
      rad2 := topradius+((bottomradius-topradius)/stacks)*(j+1);
      V1[1]:= phi;//   -h+(j*stepv);//  SinP;
      V2[1]:= phi2;//-h+((j+1)*stepv);//  SinP2;
      vTexCoord0:=1-j*vTexFactor;
      vTexCoord1:=1-(j+1)*vTexFactor;

      pv1 := -1;
      cv1 := -1;
      pv2 := -1;
      cv2 := -1;
      for i:=0 to Slices do begin

         SinCos(Theta, SinT, CosT);
         V1[0]:=Sint * rad1;//SinT;
         V2[0]:=Sint * rad2;// SinT;
         V1[2]:=Cost * rad1;//CosT;
         V2[2]:=cost * rad2;//CosT;

         uTexCoord:=i*uTexFactor;
         mo.TexCoords.Add(uTexCoord, vTexCoord0);
         mo.Normals.Add(v1);
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);

         mo.TexCoords.Add(uTexCoord, vTexCoord1);
         mo.Normals.Add(v2);
         pv2 := cv2;
         cv2 := mo.Vertices.Add(v2);

         if pv1 <> -1 then
         begin
            fg.Add(pv1);
            fg.Add(cv2);
            fg.add(cv1);
            fg.Add(cv2);
            fg.Add(pv1);
            fg.add(pv2);
         end;

         Theta:=Theta+StepH;
      end;
      Phi:=Phi2;
      Phi2:=Phi2 - StepV;
   end;
  end;

  if cybottom in parts then
  begin
      mo.TexCoords.add(0.5,0.5);
      fv1 := mo.Vertices.add(0,-h,0);
      N1:=YVector;
      V1[1]:=-h;
      Theta:=AngStop;
      cv1 := -1;
      pv1 := -1;
      for I:=0 to Slices do begin
         SinCos(Theta, SinT, CosT);
         V1[0]:=SinT*bottomradius;
         V1[2]:=CosT*bottomradius;
         mo.TexCoords.Add(SinT*0.5+0.5, CosT*0.5+0.5);
         mo.Normals.add(n1);
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);
         if pv1 <> -1 then
         begin
            fg.Add(fv1);
            fg.Add(pv1);
            fg.add(cv1);
         end;
         Theta:=Theta - StepH;
      end;
  end;

  if cytop in parts then
  begin
      mo.TexCoords.add(0.5,0.5);
      fv1 := mo.Vertices.add(0,h,0);
      N1:=YVector;
      V1[1]:=h;
      Theta:=AngStop;
      cv1 := -1;
      pv1 := -1;
      for I:=0 to Slices do begin
         SinCos(Theta, SinT, CosT);
         V1[0]:=SinT*topradius;
         V1[2]:=CosT*topradius;
         mo.TexCoords.Add(SinT*0.5+0.5, CosT*0.5+0.5);
         mo.Normals.add(n1);
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);
         if pv1 <> -1 then
         begin
            fg.Add(fv1);
            fg.Add(cv1);
            fg.add(pv1);
         end;
         Theta:=Theta - StepH;
      end;
  end;

END;

procedure BuildMeshDodecahedron(mo:tmeshobject);
const
   A = 1.61803398875*0.3; // (Sqrt(5)+1)/2
   B = 0.61803398875*0.3; // (Sqrt(5)-1)/2
   C = 1*0.3;
const
   vertices : packed array [0..19] of TAffineVector =
      ((-A,  0,  B), (-A,  0, -B), ( A,  0, -B), (A,  0,  B),
       ( B, -A,  0), (-B, -A,  0), (-B,  A,  0), (B,  A,  0),
       ( 0,  B, -A), ( 0, -B, -A), ( 0, -B,  A), (0,  B,  A),
       (-C, -C,  C), (-C, -C, -C), ( C, -C, -C), (C, -C,  C),
       (-C,  C,  C), (-C,  C, -C), ( C,  C, -C), (C,  C,  C));

   polygons : packed array [0..11] of packed array [0..4] of Byte =
      (( 0, 12, 10, 11, 16),  ( 1, 17,  8,  9, 13),
       ( 2, 14,  9,  8, 18),  ( 3, 19, 11, 10, 15),
       ( 4, 14,  2,  3, 15),  ( 5, 12,  0,  1, 13),
       ( 6, 17,  1,  0, 16),  ( 7, 19,  3,  2, 18),
       ( 8, 17,  6,  7, 18),  ( 9, 14,  4,  5, 13),
       (10, 12,  5,  4, 15),  (11, 19,  7,  6, 16));

var
  fg : TFGVertexIndexList;
  fv1,cv1,pv1 : integer;
  fv2,cv2,pv2 : integer;

   i, j : Integer;
   n : TAffineVector;
   faceIndices : PByteArray;
begin
   FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);

   for i:=0 to 11 do begin
      faceIndices:=@polygons[i, 0];

      n:=CalcPlaneNormal(vertices[faceIndices[0]],
                         vertices[faceIndices[1]],
                         vertices[faceIndices[2]]);
      mo.Normals.Add(n);
      for j:=0 to 4 do
      begin
         cv1 := mo.Vertices.add(vertices[faceIndices[j]]);  ;//glVertex3fv(@vertices[faceIndices[j]]);
         if j = 0 then fv1 := cv1;
      end;
      fg.add(fv1);
      fg.add(fv1+1);
      fg.add(fv1+2);

      fg.add(fv1);
      fg.add(fv1+2);
      fg.add(fv1+3);

      fg.add(fv1);
      fg.add(fv1+3);
      fg.add(fv1+4);

      fg.add(fv1);
      fg.add(fv1+1);
      fg.add(fv1+2);

   end;
end;


procedure BuildMeshDisk(mo:tmeshobject;innerradius,outerradius,startangle,sweepangle:single;loops,slices:integer);
var
  fg : TFGVertexIndexList;
  fv1,cv1,pv1 : integer;
  fv2,cv2,pv2 : integer;
  rad1,rad2,angstart,angstop,steph,stepv : extended;
  uvrad1,uvrad2,uvstep :extended;
  i,j:integer;
  V1, V2, N1 : TAffineVector;
  SinP, CosP, SinP2, CosP2, SinT, CosT, Phi, Phi2, Theta: Extended;
begin

   FG := TFGVertexIndexList.CreateOwned(mo.FaceGroups);

   AngStart:=DegToRad(startangle);
   AngStop:=DegToRad(sweepangle);
   StepH:=(AngStop - AngStart) / Slices;
   StepV:=(outerradius-innerradius) / loops;
   uvstep := 0.5 / loops;

   N1:=YVector;
//   n1[1] := -n1[1];
   V1[2]:=0;
   V2[2]:=0;
   rad1 := outerradius;
   for J:=0 to loops-1 do
   begin
    rad2 := rad1-stepv;
    Theta:=AngStop;
    cv1 := -1;
    pv1 := -1;
    cv2 := -1;
    pv2 := -1;
    uvrad1 := (uvstep)*j;
    uvrad2 := (uvstep)*(j+1);

    for i := 0 to slices do
    begin
         SinCos(Theta, SinT, CosT);
         V1[1]:=SinT*rad1;
         V1[0]:=CosT*rad1;
         V2[1]:=SinT*rad2;
         V2[0]:=CosT*rad2;
//         mo.TexCoords.Add(SinT*0.5+0.5, CosT*0.5+0.5);
         mo.TexCoords.Add(SinT*uvrad1+0.5, CosT*uvrad1+0.5);
         mo.TexCoords.Add(SinT*uvrad2+0.5, CosT*uvrad2+0.5);
         mo.Normals.add(n1);
         mo.Normals.add(n1);
         
         pv1 := cv1;
         cv1 := mo.Vertices.Add(v1);
         pv2 := cv2;
         cv2 := mo.Vertices.Add(v2);
         if pv1 <> -1 then
         begin
            fg.Add(pv1);
            fg.Add(cv2);
            fg.add(cv1);
            fg.Add(cv2);
            fg.Add(pv1);
            fg.add(pv2);
         end;
         Theta:=Theta - StepH;
    end;
    rad1 := rad1-stepv;
   end;




end;


procedure BuildMeshTorus(mo : tmeshobject;MinorRadius,MajorRadius:single);
begin
//
end;




end.
