{******************************************************************************}
{                                                                              }
{                              GmXml.pas 0.12                                  }
{                                                                              }
{           Copyright (c) 2003 Graham Murt  - www.murtsoft.co.uk               }
{                                                                              }
{   Feel free to e-mail me with any comments, suggestions, bugs or help at:    }
{                                                                              }
{                           graham@murtsoft.co.uk                              }
{                                                                              }
{******************************************************************************}

unit GmXml;

interface

uses Classes, SysUtils;

const
  COMP_VERSION      = 0.12;

  XML_SPECIFICATION = '<?xml version="1.0"?>';

type
  TGmXmlNode = class;

  TGmXmlTokenKind = (gmXmlTagVersion, gmXmlTagOpen, gmXmlTagClose, gmXmlElement);

  TGmXmlEnumNodeEvent = procedure(Sender: TObject; ANode: TGmXmlNode) of object;

  // *** TGmXmlNodeElement ***

  TGmXmlNodeAttribute = class
  private
    FName: string;
    FValue: string;
    procedure SetName(AValue: string);
    procedure SetValue(AValue: string);
  public
    property Name: string read FName write SetName;
    property Value: string read FValue write SetValue;
  end;

//------------------------------------------------------------------------------

  TGmXmlNodeList = class;

  // *** TGmXmlNode ***

  TGmXmlNode = class(TPersistent)
  private
    FChildren: TGmXmlNodeList;
    FElement: TGmXmlNodeAttribute;
    FName: string;
    FParent: TGmXmlNode;

    FValue: string;
    // events...
    function GetIsLeafNode: Boolean;
    function GetAsBoolean: Boolean;
    function GetAsFloat: Extended;
    function GetAsInteger: integer;
    function GetAsString: string;
    procedure SetAsBoolean(const Value: Boolean);
    procedure SetAsFloat(const Value: Extended);
    procedure SetAsInteger(const Value: integer);
    procedure SetAsString(const Value: string);
    procedure SetName(Value: string); virtual;
  public
    constructor Create(AParentNode: TGmXmlNode); virtual;
    destructor Destroy; override;
    function CloseTag: string;
    function Level: integer;
    function OpenTag: string;
    function GetNamedItemAttribute(ItemAttribute: string): string;
    function AttributesCount: integer;
    procedure EnumerateNodes(ACallback: TGmXmlEnumNodeEvent);
    property AsString: string read GetAsString write SetAsString;
    property AsBoolean: Boolean read GetAsBoolean write SetAsBoolean;
    property AsFloat: Extended read GetAsFloat write SetAsFloat;
    property AsInteger: integer read GetAsInteger write SetAsInteger;
    property Attribute: TGmXmlNodeAttribute read FElement write FElement;
    property Children: TGmXmlNodeList read FChildren;
    property IsLeafNode: Boolean read GetIsLeafNode;
    property Name: string read FName write SetName;
    property Parent: TGmXmlNode read FParent;
  end;

//------------------------------------------------------------------------------

  // *** TGmXmlNodeList ***

  TGmXmlNodeList = class
  private
    FCurrentNode: TGmXmlNode;
    FList: TList;
    function GetCount: integer;
    function GetVersion: string;
    function GetNode(index: integer): TGmXmlNode;
    function GetNodeByName(AName: string): TGmXmlNode;
    procedure SetNodeByName(AName: string; ANode: TGmXmlNode);
    function GetRoot: TGmXmlNode;
    procedure AddNode(ANode: TGmXmlNode);
    procedure SetNode(index: integer; const Value: TGmXmlNode);
  public
    constructor Create(AParent: TGmXmlNode);
    destructor Destroy; override;
    function AddLeaf(AName: string): TGmXmlNode;
    function AddOpenTag(AName: string): TGmXmlNode;
    procedure AddCloseTag;
    procedure NextNode;
    procedure Clear;
    procedure AddXMLString(AXMLString: string);
    property Count: integer read GetCount;
    property CurrentNode: TGmXmlNode read FCurrentNode write FCurrentNode;
    property Node[index: integer]: TGmXmlNode read GetNode write SetNode; default;
    property NodeByName[AName: string]: TGmXmlNode read GetNodeByName write SetNodeByName;
    function GetNodeByName2(AName: string): TGmXmlNode;
    property Root: TGmXmlNode read GetRoot;
    property Version: string read GetVersion;
  end;

//------------------------------------------------------------------------------

  // *** TGmXML ***

  TGmXML = class(TComponent)
  private
    FAutoIndent: Boolean;
    FNodes: TGmXmlNodeList;
    //function GetAbout: string;
    function GetCR: string;
    function GetIndent(ALevel: integer): string;
    function GetText: string;
    procedure SetAsText(Value: string);
    //procedure SetAbout(Value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadFromFile(AFileName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToFile(AFilename: string);
    procedure SaveToStream(Stream: TStream);
    property AutoIndent: Boolean read FAutoIndent write FAutoIndent default True;
    property Nodes: TGmXmlNodeList read FNodes;
    property Text: string read GetText write SetAsText;
    function GetNodeByName(AName: string): TGmXmlNode;
    function GetNodeByPath(APath: string): TGmXmlNode;
    function GetStringByPath(APath: string): string;
    procedure NodeToString(var AXml: string; ANode: TGmXmlNode);
  published
    //property About: string read GetAbout write SetAbout;
  end;

function TokTok(sep: string; var s: string): string;

//------------------------------------------------------------------------------

//procedure Register;

implementation

//------------------------------------------------------------------------------

// *** Unit Functions ***

procedure ReplaceText(var AText: string; AFind, AReplace: string);
var
  Index: integer;
begin
  while Pos(AFind, AText) <> 0 do
  begin
    Index := Pos(AFind, AText);
    Delete(AText, Index, Length(AFind));
    Insert(AReplace, AText, Index);
  end;
end;

function GetNextToken(AXml: string; var CursorPos: integer): string;
begin
  Result := '';
  if AXml[CursorPos] = '<' then
  begin
    // token is an XML tag...
    repeat
      Result := Result + AXml[CursorPos];
      Inc(CursorPos);
    until AXml[CursorPos-1] = '>';
  end
  else
  begin
    // token is an XML value
    repeat
      Result := Result + AXml[CursorPos];
      Inc(CursorPos);
    until AXml[CursorPos] = '<';
  end;
end;

function GetTokenKind(AToken: string): TGmXmlTokenKind;
begin
  Result := gmXmlElement;
  if Length(AToken) < 2 then Exit;
  if AToken[1] = '<' then
  begin
    if AToken[2] = '/' then
      Result := gmXmlTagClose
    else
    if AToken[2] = '?' then
      Result := gmXmlTagVersion
    else
      Result := gmXmlTagOpen;
  end;
end;

function TokTok(sep: string; var s: string): string;
  function IsOneOf(c, s: string): Boolean;
  var
    iTmp: integer;
  begin
    Result:= False;
    for iTmp:= 1 to Length(s) do begin
      if c = Copy(s, iTmp, 1) then begin
        Result:= True;
        Exit;
        end;
      end;
  end;
var
  c, t: string;
begin
  if s = '' then begin
    Result:= s;
    Exit;
    end;
  c:= Copy(s, 1, 1);
  while IsOneOf(c, sep) do begin
    s:= Copy(s, 2, Length(s) - 1);
    c:= Copy(s, 1, 1);
    end;
  t:= '';
  while (not IsOneOf(c, sep)) and (s <> '') do begin
    t:= t + c;
    s:= Copy(s, 2, length(s)-1);
    c:= Copy(s, 1, 1);
    end;
  Result:= t;
end;

//------------------------------------------------------------------------------

// *** TGmXmlNodeElement ***

procedure TGmXmlNodeAttribute.SetName(AValue: string);
begin
  FName := Trim(AValue);
end;

procedure TGmXmlNodeAttribute.SetValue(AValue: string);
begin
  FValue := Trim(AValue);
end;

//------------------------------------------------------------------------------

// *** TGmXmlNode ***

constructor TGmXmlNode.Create(AParentNode: TGmXmlNode);
begin
  inherited Create;
  FChildren := TGmXmlNodeList.Create(Self);
  FParent := AParentNode;
  FElement := TGmXmlNodeAttribute.Create;
  FElement.Name := '';
  FElement.Value := '';
end;

destructor TGmXmlNode.Destroy;
begin
  FElement.Free;
  FChildren.Free;
  inherited Destroy;
end;

function TGmXmlNode.CloseTag: string;
begin
  Result := Format('<%s>',[Name]);
  Insert('/', Result, 2);
end;

function TGmXmlNode.Level: integer;
var
  AParent: TGmXmlNode;
begin
  AParent := Parent;
  Result := 0;
  while AParent <> nil do
  begin
    AParent := AParent.Parent;
    Inc(Result);
  end;
end;

function TGmXmlNode.OpenTag: string;
var
  nnn, vvv, rrr: string;
begin
  if FElement.Name = '' then
    Result := Format('<%s>',[Name])
  else begin
    rrr:= '<' + Name;
    nnn:= FElement.Name;
    vvv:= FElement.Value;
    if Pos('|', nnn) > 0
      then begin
        while Pos('|', nnn) > 0 do
          rrr:= rrr + ' ' + TokTok('|', nnn) + '="' + TokTok('|', vvv) + '"';
        end
      else rrr:= rrr + ' ' + nnn + '="' + vvv + '"';
    rrr:= rrr + '>';
    Result:= rrr;
    end;
end;

function TGmXmlNode.GetNamedItemAttribute(ItemAttribute: string): string;
var
  nnn, vvv, nn, vv, rrr: string;
begin
  rrr:= '';
  nnn:= FElement.Name;
  vvv:= FElement.Value;
  while nnn <> '' do begin
    nn:= TokTok('|', nnn);
    vv:= TokTok('|', vvv);
    if nn = ItemAttribute then rrr:= vv;
    end;
  Result:= rrr;
end;

function TGmXmlNode.AttributesCount: integer;
var
  nnn: string;
  i: integer;
begin
  i:= 0;
  nnn:= FElement.Name;
  while Pos('|', nnn) > 0 do begin
    i:= i + 1;
    nnn[Pos('|', nnn)] := '0';
    end;
  Result:= i;
end;


procedure TGmXmlNode.EnumerateNodes(ACallback: TGmXmlEnumNodeEvent);
var
  ICount: integer;
begin
  for ICount := 0 to FChildren.Count-1 do
  begin
    if Assigned(ACallback) then ACallback(Self, FChildren[ICount]);
  end;
end;

function TGmXmlNode.GetAsBoolean: Boolean;
begin
  Result := Boolean(StrToInt(FValue));
end;

function TGmXmlNode.GetAsFloat: Extended;
begin
  Result := StrToFloat(FValue);
end;

function TGmXmlNode.GetAsInteger: integer;
begin
  Result := StrToInt(FValue);
end;

function TGmXmlNode.GetAsString: string;
begin
  Result := FValue;
end;

procedure TGmXmlNode.SetAsBoolean(const Value: Boolean);
begin
  FValue := IntToStr(Ord(Value));
end;

procedure TGmXmlNode.SetAsFloat(const Value: Extended);
begin
  FValue := FloatToStr(Value);
end;

procedure TGmXmlNode.SetAsInteger(const Value: integer);
begin
  FValue := IntToStr(Value);
end;

procedure TGmXmlNode.SetAsString(const Value: string);
begin
  FValue := Value;
end;

function TGmXmlNode.GetIsLeafNode: Boolean;
begin
  Result := FChildren.Count = 0;
end;

procedure TGmXmlNode.SetName(Value: string);
var
  AElement, nnn, vvv, rrr, nn, vv: string;
begin
  FName := Value;
  if FName[1] = '<' then Delete(FName, 1, 1);
  if FName[Length(FName)] = '>' then Delete(FName, Length(FName), 1);
  Trim(FName);

  // extract element if one exists...
  if Pos('=', FName) <> 0 then
  begin
    rrr := Copy(FName, Pos(' ', FName), Length(FName));
    FName := Copy(FName, 1, Pos(' ', FName)-1);
    while Pos('=', rrr) > 0 do begin
      nn:= Copy(rrr, 1, Pos('=', rrr)-1);
      nn:= Trim(nn);
      if nnn = ''
        then nnn:= nn
        else nnn:= nnn + '|' + nn;
      rrr:= Copy(rrr, Pos('"', rrr)+1, Length(rrr));
      vv:= Copy(rrr, 1, Pos('"', rrr)-1);
      vv:= Trim(vv);
      if vvv = ''
        then vvv:= vv
        else vvv:= vvv + '|' + vv;
      rrr:= Copy(rrr, Pos('"', rrr)+1, Length(rrr));
      rrr:= Trim(rrr);
      end;

    nnn:= Trim(nnn);
    vvv:= Trim(vvv);
    FElement.Name:= nnn;
    ReplaceText(vvv, '"', '');
    FElement.Value:= vvv;
  end;
end;


//------------------------------------------------------------------------------

// *** TGmXmlNodeList ***

constructor TGmXmlNodeList.Create(AParent: TGmXmlNode);
begin
  inherited Create;
  FList := TList.Create;
  FCurrentNode := AParent;
end;

destructor TGmXmlNodeList.Destroy;
var
  ICount: integer;
begin
  for ICount := Count-1 downto 0 do
    Node[ICount].Free;
  FList.Free;
  inherited Destroy;
end;

function TGmXmlNodeList.AddLeaf(AName: string): TGmXmlNode;
begin
  Result := AddOpenTag(AName);
  AddCloseTag;
end;

function TGmXmlNodeList.AddOpenTag(AName: string): TGmXmlNode;
begin
  Result := TGmXmlNode.Create(FCurrentNode);
  Result.Name := AName;
  if FCurrentNode = nil then
    AddNode(Result)
  else
    FCurrentNode.Children.AddNode(Result);

  FCurrentNode := Result;
end;

procedure TGmXmlNodeList.AddCloseTag;
begin
  FCurrentNode := FCurrentNode.Parent;
end;

procedure TGmXmlNodeList.NextNode;
var
  AIndex: integer;
begin
  AIndex := FList.IndexOf(FCurrentNode);
  if AIndex < FList.Count then
  FCurrentNode := TGmXmlNode(FList[AIndex]);
end;

procedure TGmXmlNodeList.Clear;
var
  ICount: integer;
begin
  for ICount := 0 to FList.Count-1 do
  begin
    Node[ICount].Free;
    Node[ICount] := nil;
  end;
  FList.Clear;
  FCurrentNode := nil;
end;

function TGmXmlNodeList.GetCount: integer;
begin
  Result := FList.Count;
end;

function TGmXmlNodeList.GetVersion: string;
begin
  Result := '<?xml version="1.0"?>';
end;

function TGmXmlNodeList.GetNode(index: integer): TGmXmlNode;
begin
  Result := TGmXmlNode(FList[index]);
end;


function TGmXmlNodeList.GetNodeByName(AName: string): TGmXmlNode;
var
  ICount: integer;
begin
  Result := nil;
  for ICount := 0 to Count-1 do
  begin
    if Node[ICount].Name = AName then
    begin
      Result := Node[ICount];
      Exit;
    end;
  end;
end;

function TGmXmlNodeList.GetNodeByName2(AName: string): TGmXmlNode;

  procedure Recursion(AName: string; ANode: TGmXmlNode; var ResultNode: TGmXmlNode);
  var
    ICount: integer;
  begin
    if ResultNode <> nil then exit;
    if ANode.Name = AName
      then ResultNode:= ANode
      else begin
        if ANode.FChildren.Count > 0 then
          for ICount:= 0 to ANode.FChildren.Count - 1 do
            Recursion(AName, ANode.Children.Node[ICount], ResultNode);
        end;
  end;

var
  ICount: integer;
begin
  Result := nil;
  Recursion(AName, GetRoot, Result);
end;

procedure TGmXmlNodeList.SetNodeByName(AName: string; ANode: TGmXmlNode);
var
  ICount: integer;
begin
  for ICount := 0 to Count-1 do
  begin
    if Node[ICount].Name = AName then
    begin
      Node[ICount] := ANode;
      Exit;
    end;
  end;
end;

function TGmXmlNodeList.GetRoot: TGmXmlNode;
begin
  Result := nil;
  if Count > 0 then Result := Node[0];
end;

procedure TGmXmlNodeList.AddNode(ANode: TGmXmlNode);
begin
  FList.Add(ANode);
end;

procedure TGmXmlNodeList.SetNode(index: integer; const Value: TGmXmlNode);
begin
  FList[index] := Value;
end;

procedure TGmXmlNodeList.AddXMLString(AXMLString: string);
var
  ALines: TStringList;
  ACursor: integer;
  ICount: integer;
  TokenBefore,
  TokenAfter: TGmXmlTokenKind;
  AText: string;
begin
  AText := AXMLString;
  ReplaceText(AText, #13, '');
  ReplaceText(AText, #10, '');

  ACursor := 1;
  ALines := TStringList.Create;
  try
    // first we need to identify valid tags and data and add them to a list...
    while ACursor < Length(AText) do
    begin
      ALines.Add(GetNextToken(AText, ACursor));
    end;
    // now we need to remove any indentation for internal storage...
    for ICount := ALines.Count-1 downto 1 do
    begin
      if GetTokenKind(ALines[ICount]) = gmXmlElement then
      begin
        TokenBefore := GetTokenKind(ALines[ICount-1]);
        TokenAfter  := GetTokenKind(ALines[ICount+1]);
        if not ((TokenBefore = gmXmlTagOpen) and (TokenAfter = gmXmlTagClose)) then
          ALines.Delete(ICount);
      end;
    end;
    // build the node list...
    for ICount := 0 to ALines.Count-1 do
    begin
      case GetTokenKind(ALines[ICount]) of
        gmXmlTagVersion: ;
        gmXmlTagOpen   : AddOpenTag(ALines[ICount]);
        gmXmlElement   : FCurrentNode.AsString := ALines[ICount];
        gmXmlTagClose  : AddCloseTag;
      end;
    end;
  finally
    ALines.Free;
  end;
end;

//------------------------------------------------------------------------------

// *** TGmXml ***

constructor TGmXml.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FNodes := TGmXmlNodeList.Create(nil);
  FAutoIndent := True;
end;

destructor TGmXml.Destroy;
begin
  FNodes.Free;
  inherited Destroy;
end;

//function TGmXml.GetAbout: string;
//begin
//  Result := Format('%s v%f', [ClassName, COMP_VERSION]);
//end;

function TGmXml.GetCR: string;
begin
  if FAutoIndent then
    Result := #13#10
  else
    Result := '';
end;

function TGmXml.GetIndent(ALevel: integer): string;
var
  ICount: integer;
begin
  Result := '';
  if FAutoIndent = False then Exit;
  for ICount := 1 to ALevel do
    Result := Result + '  ';
end;

procedure TGmXml.NodeToString(var AXml: string; ANode: TGmXmlNode);
var
  ICount: integer;
begin
  if ANode.IsLeafNode then
  begin
    AXml := AXml + GetIndent(ANode.Level) + ANode.OpenTag + ANode.AsString + ANode.CloseTag + GetCR;
  end
  else
  begin
    AXml := AXml + GetIndent(ANode.Level)+ANode.OpenTag + GetCR;
    for ICount := 0 to ANode.FChildren.Count-1 do
      NodeToString(AXml, ANode.Children.Node[ICount]);
    AXml := AXml + GetIndent(ANode.Level)+ANode.CloseTag + GetCR;
  end;
end;

function TGmXml.GetText: string;
var
  ICount: integer;
begin
  Result := '';
  if FNodes.Count = 0 then Exit;
  Result := FNodes.Version+ GetCR;
  for ICount := 0 to FNodes.Count-1 do
    NodeToString(Result, FNodes.Node[ICount]);
end;

procedure TGmXml.SetAsText(Value: string);
var
  ALines: TStringList;
  ACursor: integer;
  ICount: integer;
  TokenBefore,
  TokenAfter: TGmXmlTokenKind;
  AText: string;
begin
  AText := Value;
  //ReplaceText(AText, #13, '');
  //ReplaceText(AText, #10, '');
  ReplaceText(AText, #13 + #10, '');

  ACursor := 1;
  ALines := TStringList.Create;
  try
    // first we need to identify valid tags and data and add them to a list...
    while ACursor < Length(AText) do
    begin
      ALines.Add(GetNextToken(AText, ACursor));
    end;
    // now we need to remove any indentation for internal storage...
    for ICount := ALines.Count-1 downto 1 do
    begin
      if GetTokenKind(ALines[ICount]) = gmXmlElement then
      begin
        TokenBefore := GetTokenKind(ALines[ICount-1]);
        TokenAfter  := GetTokenKind(ALines[ICount+1]);
        if not ((TokenBefore = gmXmlTagOpen) and (TokenAfter = gmXmlTagClose)) then
          ALines.Delete(ICount);
      end;
    end;
    // build the node list...
    FNodes.Clear;
    for ICount := 0 to ALines.Count-1 do
    begin
      case GetTokenKind(ALines[ICount]) of
        gmXmlTagVersion: ;
        gmXmlTagOpen   : FNodes.AddOpenTag(ALines[ICount]);
        gmXmlElement   : FNodes.CurrentNode.AsString := ALines[ICount];
        gmXmlTagClose  : FNodes.AddCloseTag;
      end;
    end;
  finally
    ALines.Free;
  end;
end;

//procedure TGmXml.SetAbout(Value: string);
//begin
  // does nothing... (only needed to display property in Object Inspector)
//end;

procedure TGmXml.LoadFromFile(AFileName: string);
var
  AStream: TMemoryStream;
begin
  AStream := TMemoryStream.Create;
  try
    AStream.LoadFromFile(AFileName);
    AStream.Seek(0, soFromBeginning);
    LoadFromStream(AStream);
  finally
    AStream.Free;
  end;
end;

procedure TGmXml.LoadFromStream(Stream: TStream);
var
  ALines: TStringList;
  a, b: string;
  i: integer;
begin
  if Stream.Size = 0 then Exit;
  ALines := TStringList.Create;
  try
    ALines.LoadFromStream(Stream);
    //Text := ALines.Text;
    a:= '';
    if ALines.Count > 0 then
     for i:= 0 to ALines.Count - 1 do begin
       b:= ALines[i];
       ReplaceText(b, #13 + #10, '');
       a:= a + b;
       end;
    Text:= a;
  finally
    ALines.Free;
  end;
end;

procedure TGmXml.SaveToFile(AFilename: string);
var
  AStream: TMemoryStream;
begin
  AStream := TMemoryStream.Create;
  try
    SaveToStream(AStream);
    AStream.SaveToFile(AFilename);
  finally
    AStream.Free;
  end;
end;

procedure TGmXml.SaveToStream(Stream: TStream);
var
  ALines: TStringList;
begin
  ALines := TStringList.Create;
  try
    ALines.Text := Text;
    ALines.SaveToStream(Stream);
  finally
    ALines.Free;
  end;
end;

function TGmXml.GetNodeByName(AName: string): TGmXmlNode;

  procedure Recursion(AName: string; ANode: TGmXmlNode; var ResultNode: TGmXmlNode);
  var
    ICount: integer;
  begin
    if ResultNode <> nil then exit;
    if ANode.Name = AName
      then ResultNode:= ANode
      else begin
        if ANode.FChildren.Count > 0 then
          for ICount:= 0 to ANode.FChildren.Count - 1 do
            Recursion(AName, ANode.Children.Node[ICount], ResultNode);
        end;
  end;

begin
  Result := nil;
  Recursion(AName, Nodes.GetRoot, Result);
end;

function TGmXml.GetNodeByPath(APath: string): TGmXmlNode;

  procedure Recursion(APath: string; ANode: TGmXmlNode; var ResultNode: TGmXmlNode);
  var
    i, j, ICount: integer;
    ANodeChildren: TGmXmlNode;
    a, b: string;
  begin
    a:= APath;
    i:= Pos('\', a);
    if i > 0
      then begin
        for j:= 1 to i - 1 do b:= b + a[j];
        Delete(a, 1, i);
        ANodeChildren:= ANode.Children.GetNodeByName(b);
        Recursion(a, ANodeChildren, ResultNode);
        end
      else ResultNode:= ANode.Children.GetNodeByName(a);
  end;

begin
  Result := nil;
  Recursion(APath, Nodes.GetRoot, Result);
end;
                         
function TGmXml.GetStringByPath(APath: string): string;
var
  GmXmlNode: TGmXmlNode;
begin
  Result:= '';
  try GmXmlNode:= GetNodeByPath(APath) except end;
  if GmXmlNode <> nil
    then Result:= GmXmlNode.AsString
    else Result:= '';
end;


//procedure Register;
//begin
//  RegisterComponents('GmXML', [TGmXml]);
//end;

end.
